/* eslint-disable no-plusplus */
/* global Proxy, Set */

'use strict';

const


  /**
   * Checks the types passed into pluck and indexBy
   * if key is not a string, throw
   * if obj is array return itself
   * if obj is not array return object.values
   *
   * @param {Object} obj object of items
   * @param {String} key the string name of property
   *
   * @return {Array.<Object>} array of values in object
   **/
  values_from_array_or_obj = (obj, key) => {
    if ('string' !== typeof key) {
      throw new Error('key for pluck must be a string');
    }

    if (true === Array.isArray(obj)) {
      return obj;
    } else if ('object' === typeof obj) {
      return Object.values(obj);
    }

    throw new Error('object for pluck/indexBy must be array or object');
  }

  , realUnderscore = require('underscore')

  , ourUnderscore = {

    /**
     * Return array of items that are in array and NOT in array2
     *
     * @param {Array.<Object>} array array of items
     * @param {Array.<Object>} array2 array of items
     * @param {Array.<Object>} extra extra args to throw on
     *
     * @return {Array.<Object>} items unique to array
     **/
    difference: (array, array2, ...extra) => {
      if (0 !== extra.length) {
        throw new Error('too many arguments for difference');
      }
      const set = new Set(array2)
        , uniq = new Set(array)
        , values = [];

      for (const item of uniq) {
        if (false === set.has(item)) {
          values.push(item);
        }
      }
      return values;
    }

    /**
     * Return array of items that are in array and in array2
     *
     * @param {Array.<Object>} array array of items
     * @param {Array.<Object>} array2 array of items
     * @param {Array.<Object>} extra extra args to throw on
     *
     * @return {Array.<Object>} items in common between arrays
     **/
    , intersection: (array, array2, ...extra) => {
      if (0 !== extra.length) {
        throw new Error('too many arguments for intersection');
      }
      const set = new Set(array2)
        , uniq = new Set(array)
        , values = [];

      for (const item of uniq) {
        if (true === set.has(item)) {
          values.push(item);
        }
      }
      return values;
    }

    /**
     * Return array of values from key
     *
     * @param {Object} obj obj of items
     * @param {String} key the string name of property value to return
     *
     * @return {Array.<Object>} array of values for key
     **/
    , pluck: (obj, key) => {
      const values = values_from_array_or_obj(obj, key)
        , length = values.length
        , plucked = Array(length);
      for (let i = 0; i < length; i++) {
        const item = values[i];
        plucked[i] = item[key];
      }

      return plucked;
    }

    /**
     * Return hash of items that are indexed by key
     *
     * @param {Object} obj obj of items
     * @param {String} key the string name of property to index by
     *
     * @return {Object} hash of key -> item
     **/
    , indexBy: (obj, key) => {
      const values = values_from_array_or_obj(obj, key)
        , length = values.length
        , indexed = {};
      for (let i = 0; i < length; i++) {
        const item = values[i];
        indexed[item[key]] = item;
      }

      return indexed;
    },
  }

  /**
   * Return result of function call from either our or real underscore
   *
   * @param {Object} firstArgument the first argument passed into function call
   *
   * @return {Object} result of calling underscore function
   **/
  , constructor = (firstArgument) => {
    return new Proxy({}, {
      get: (obj, method) => {
        return (...args) => {
          return (ourUnderscore[method] ||
            realUnderscore[method])(firstArgument, ...args);
        };
      },
    });
  }

  /**
   * Return an object that runs ourUnderscore defined methods
   * and if those are undefined runs the underscore lib version
   *
   * @param {Object} constructor func to ensure all arguments are passed down
   * @param {Object} handler trap for getting a property value
   *
   * @return {Object} underscore object with our custom implementations
   **/
  , combinedUnderscore = new Proxy(constructor, {
    get: (obj, method) => {
      return ourUnderscore[method] || realUnderscore[method];
    },
  })

  /**
   * Go through all chained actions and run the underscore function
   *
   * @param {Object} firstData the inital data to operate on
   * @param {Array<Object>} actions array of underscore functions
   *
   * @return {Object} data after all computations are done
   **/
  , compute = (firstData, actions) => {
    const length = actions.length;
    let data = firstData;
    for (let i = 0; i < length; i++) {
      const action = actions[i]
        , method = combinedUnderscore[action.name];
      data = method ? method(data, ...action.args) :
        realUnderscore(data)[action.name](...action.args);
    }
    return data;
  }

  /**
   * chain all underscore functions and call compute
   *
   * @param {Object} firstData the inital data to operate on
   * @param {Array<Object>} actions array of underscore functions
   *
   * @return {Object} data after all computations are done
   **/
  , chain = (firstData, actions = []) => {
    return new Proxy({}, {
      get: (obj, name) => {
        if ('value' === name) {
          return () => {
            return compute(firstData, actions);
          };
        }
        return (...args) => {
          actions.push({
            name,
            args,
          });
          return chain(firstData, actions);
        };
      },
    });
  };

ourUnderscore.chain = chain;
module.exports = combinedUnderscore;
